# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8-80 compliant>
bl_addon_info = {
    'name': 'Mesh: Snap With Axis Constraints',
    'author': 'Josh Wedlake',
    'version': '0.1',
    'blender': (2, 5, 3),
    'location': 'ToolShelf Search',
    'description': 'Snap (and remove doubled vertices) with axis constraints',
    'url': 'http://tube.freefac.org',
    'category': 'Mesh'}

import bpy
from bpy.ops import *
import math


class SnapWithinAxis(bpy.types.Operator):
    '''snap/removed doubles constrained within axis'''
    bl_idname = "mesh.snap_with_axis_constraints"
    bl_label = "Snap With Axis Constraints"
    bl_options = {'REGISTER', 'UNDO'}
    snap_tolerance = bpy.props.FloatProperty(name="Snapping Tolerance", description="snap tolerance (units)",default=0.001,min=0,max=100,step=0.01,precision=4)
    snap_iterations = bpy.props.IntProperty(name="Snap Iterations", description="number of times to repeat",default=1,min=1,max=10,step=1)
    operating_axis=bpy.props.EnumProperty(items=[('AXIS','Axis:','Axis to snap within'),('XYZ','XYZ','XYZ'),('XY','XY','XY'),('YZ','YZ','YZ'),('XZ','XZ','XZ'),('X','X','X'),('Y','Y','Y'),('Z','Z','Z')],default='XY')
    snap_point_decision=bpy.props.EnumProperty(items=[('SNAP','Snapping Mode:','Method for choosing primary snap vertex'),('NVERT','Number of neighbours within tolerance','Number of neighbours within tolerance'),('MINDIST','Minimum distance to neighbours','Minimum distance to neighbours')],default='MINDIST')
    remove_doubles=bpy.props.BoolProperty(name="Remove Doubles",description="remove doubles after snapping",default=False)
    dialog_width = 250
    
    def get_vert_selection(self,obdat):
        verts_selected=set()
        for index,ivert in enumerate(obdat.verts):
            if ivert.select==True:
                verts_selected.add(index)
        return verts_selected

    def dist_between_verts(self,obdat,verta,vertb,axis):
        import math
        if axis=='X':
            vert_dist=obdat.verts[verta].co.x-obdat.verts[vertb].co.x
            return math.fabs(vert_dist)
        elif axis=='Y':
            vert_dist=obdat.verts[verta].co.y-obdat.verts[vertb].co.y
            return math.fabs(vert_dist)
        elif axis=='Z':
            vert_dist=obdat.verts[verta].co.z-obdat.verts[vertb].co.z
            return math.fabs(vert_dist)
        elif axis=='XY':
            vert_dist=obdat.verts[verta].co.xy-obdat.verts[vertb].co.xy
            vert_dist=vert_dist.length
            return vert_dist
        elif axis=='YZ':
            vert_dist=obdat.verts[verta].co.yz-obdat.verts[vertb].co.yz
            vert_dist=vert_dist.length
            return vert_dist
        elif axis=='XZ':
            vert_dist=obdat.verts[verta].co.xz-obdat.verts[vertb].co.xz
            vert_dist=vert_dist.length
            return vert_dist
        elif axis=='XYZ':
            vert_dist=obdat.verts[verta].co-obdat.verts[vertb].co
            vert_dist=vert_dist.length
            return vert_dist
        else:
            print ('bad axis!')
            return -1
    
    def make_score(self,vert_proximity_data,decision):
        if decision=='NVERT':
            score=len(vert_proximity_data)
            return score
        elif decision=='MINDIST':
            score=sum(vert_proximity_data)
            return score
        else:
            print ('bad decision mode!')
            return -1
    
    def build_vert_data(self,obdat,selection_set,vert_index,tolerance,axis,decision):
        vert_proximity_data=[]
        vert_neighbours=[]
        for index,ivert in enumerate(obdat.verts):
            if index!=vert_index and index in selection_set:
                proximity=self.dist_between_verts(obdat,vert_index,index,axis)
                if proximity<=tolerance:
                    vert_proximity_data.append(proximity)
                    vert_neighbours.append(index)
        #now vert_proximity_data contains all the indices and distances to verts within tol
        #now find a score based on sum of dists or number of verts
        if len(vert_neighbours)>0:
            score=self.make_score(vert_proximity_data,decision)
            return [vert_neighbours,score]
        else:
            #there are no neighbours within tolerance
            return -1
    
    def build_all_verts_dictionary(self,obdat,selection_set,tolerance,axis,decision):
        vert_dict=dict()        
        for index in selection_set:
            vert_data=self.build_vert_data(obdat,selection_set,index,tolerance,axis,decision)
            if vert_data!=-1:
                vert_dict[index]=vert_data
        return vert_dict
    
    def find_lowest_score_index(self,vert_dict):
        #traverses the vertex dictionary and returns the vertex id with the lowest score
        #dict is in the form {vert id: [[neighbours...],score],...}
        #invert the dictionary
        inv_vert_dict = dict(map(lambda item: (item[1][1],item[0]),vert_dict.items()))
        #find the minimum score's vertex index in the inverted dictionray
        min_score_index = inv_vert_dict[min(inv_vert_dict.keys())]
        #return the vertex index
        return min_score_index
    
    def snap_neighbours_to_point(self,obdat,vert_index,neighbour_list,axis):
        for neighbour_vert_index in neighbour_list:
            if 'X' in axis:
                obdat.verts[neighbour_vert_index].co.x=obdat.verts[vert_index].co.x
            if 'Y' in axis:
                obdat.verts[neighbour_vert_index].co.y=obdat.verts[vert_index].co.y
            if 'Z' in axis:
                obdat.verts[neighbour_vert_index].co.z=obdat.verts[vert_index].co.z
    
    def remove_processed_data(self,vert_dict,index):
        #this function removes all the neighbours of index
        #then it removes index
        for index_to_delete in vert_dict[index][0]:
            if index_to_delete in vert_dict.keys():
                del(vert_dict[index_to_delete])
        del(vert_dict[index])
        return vert_dict 
    
    def invoke(self, context, event):
        wm = context.manager
        wm.invoke_props_dialog(self, self.dialog_width)
        return {'RUNNING_MODAL'}
    
    def execute(self, context):
        import bpy
        bpy.context.tool_settings.mesh_selection_mode=[True,False,False]
        if context.mode=='EDIT_MESH':
            bpy.ops.object.editmode_toggle()
        #create a selection set
        selection_set=self.get_vert_selection(bpy.context.active_object.data)
        for iter_run in range(0,self.properties.snap_iterations):
            #generate the dictionary
            #make sure any verts which have no neighbours within tolerance are not included
            vert_dict=self.build_all_verts_dictionary(bpy.context.active_object.data,selection_set,self.properties.snap_tolerance,self.properties.operating_axis,self.properties.snap_point_decision)
            #while there are still keys left in the dictionary
            while len(vert_dict.keys())>0:
                #find the lowest score's index
                min_score_index=self.find_lowest_score_index(vert_dict)
                #snap neighbours to the lowest scoring vertex
                self.snap_neighbours_to_point(bpy.context.active_object.data,min_score_index,vert_dict[min_score_index][0],self.properties.operating_axis)
                #remove the lowest scoring vertex AND all of the neighbours which have been repositioned
                vert_dict=self.remove_processed_data(vert_dict,min_score_index)
        bpy.ops.object.editmode_toggle()
        if self.properties.remove_doubles:
            bpy.ops.mesh.remove_doubles(limit=0)

        return {'FINISHED'}

    def draw(self, context):
        layout = self.layout
        props = self.properties

        layout.prop(props, "snap_tolerance")
        layout.prop(props, "snap_iterations")
        layout.prop(props, "operating_axis")
        layout.prop(props, "snap_point_decision")
        layout.prop(props, "remove_doubles")


specials_classes = [SnapWithinAxis]

classes=list(set(specials_classes))

def gen(c):
    def menu_func(self, context):
        self.layout.operator(c.bl_idname, c.bl_label)
    return(menu_func)

def register():
    #register = bpy.types.register
    #for c in classes:
    #        register(c)
    for c in specials_classes:
        bpy.types.VIEW3D_MT_edit_mesh_specials.append(gen(c))

def unregister():
    #unregister = bpy.types.unregister
    #for c in classes:
    #        unregister(c)
    for c in specials_classes:
        bpy.types.VIEW3D_MT_edit_mesh_specials.remove(gen(c)) 

if __name__ == "__main__":
    unregister()
    register()
